"""FieldEdge class/property helpers
"""

import json
import logging
import os
from time import time

PROPERTY_CACHE_DEFAULT = int(os.getenv('PROPERTY_CACHE_DEFAULT', 5))
VERBOSE_DEBUG = str(os.getenv('VERBOSE_DEBUG', False)).lower() == 'true'

_log = logging.getLogger(__name__)


def snake_to_camel(snake_str: str) -> str:
    """Converts a snake_case string to camelCase."""
    words = snake_str.split('_')
    return words[0] + ''.join(w.title() for w in words[1:])


def camel_to_snake(camel_str: str) -> str:
    """Converts a camelCase string to snake_case."""
    pattern = re.compile(r'(?<!^)(?=[A-Z])')
    return pattern.sub('_', camel_str).lower()


def cache_valid(ref_time: int,
                max_age: int = PROPERTY_CACHE_DEFAULT,
                tag: str = 'value',
                ) -> bool:
    """Determines if cached property value is younger than the threshold.
    
    `PROPERTY_CACHE_DEFAULT` = 5 seconds. Can be overridden as an environment
    variable.
    Many FieldEdge Class properties are derived from *slow* operations but may
    be queried in rapid succession and can be inter-dependent. Caching reduces
    query time for such values.
    
    Args:
        ref: The reference time (seconds) of the previously cached value
            (typically a private property held in a dictionary)
        max_age: The maximum age of the cached value in seconds.
        tag: The name of the property (used for debug purposes).
    
    Returns:
        False is the cache is stale and a new value should be queried from the
            raw resource.

    """
    if not isinstance(ref_time, int):
        if VERBOSE_DEBUG:
            _log.debug(f'No cached timestamp for {tag}')
        return False
    cache_age = int(time()) - ref_time
    if cache_age > max_age:
        if VERBOSE_DEBUG:
            _log.debug(f'Cached {tag} only {cache_age} seconds old'
                        f' (cache = {max_age}s)')
        return False
    _log.debug(f'Using cached {tag} ({cache_age} seconds)')
    return True


def get_class_properties(cls_instance: object) -> dict:
    """Returns non-hidden, non-callable properties/values of a Class instance.
    
    Args:
        cls_instance: The Class instance whose properties/values will be derived
    
    Returns:
        A dictionary as { 'property_name': 'property_value' }
        
    Raises:
        ValueError if cls_instance does not have a dir method.
        
    """
    if not dir(cls_instance):
        raise ValueError('Invalid cls_instanceect must have dir method')
    props = {}
    props_list = [a for a in dir(cls_instance)
                  if not a.startswith(('_', 'properties')) and
                  not callable(getattr(cls_instance, a))]
    for prop in props_list:
        props[prop] = getattr(cls_instance, prop)
    return props


def tag_class_properties(cls,
                         tag: str = None,
                         json: bool = True,
                         ignore: 'list[str]' = ['exposed_properties'],
                         ) -> dict:
    """Retrieves `@property`s from a class and tags them with a routing prefix.
    
    The dictionary items describe `read_only` and `config` items which each
    consist of a list of strings for the property names.
    Each property name is optionally prefixed by a `tag` for the class
    e.g. `modem_manufacturer`.
    A reserved name `exposed_properties` is ignored by default, but a list of
    values to ignore is supported.
    
    TODO: check `vars` vs `dir` use for robustness.

    Args:
        cls: The class to fetch properties from (`fset`)
        tag: A prefex to apply for identification/routing
        json: Property names are expressed in camelCase
        ignore: An optional filter of property names to ignore
    
    Returns:
        `{ 'config': [<str>], 'read_only': [<str>]}`

    """
    rw = [attr for attr, value in vars(cls).items()
          if isinstance(value, property) and
          value.fset is not None and
          attr not in ignore]
    ro = [attr for attr, value in vars(cls).items()
          if isinstance(value, property) and
          value.fset is None and
          attr not in ignore]
    if tag is not None:
        for i, prop in enumerate(rw):
            rw[i] = f'{tag}_{prop}'
        for i, prop in enumerate(ro):
            ro[i] = f'{tag}_{prop}'
    tagged = { 'config': rw, 'read_only': ro }
    if not json:
        return tagged
    json_tagged = {}
    for key, value in tagged.items():
        json_list = []
        for item in value:
            json_list.append(snake_to_camel(item))
        json_tagged[snake_to_camel(key)] = json_list
    return json_tagged


def untag_class_property(tagged_property: str,
                   tag: str = None,
                   include_tag: bool = False,
                   ) -> 'str|tuple[str, str]':
    """Reverts a JSON-format tagged property to its PEP representation.
    
    Expects a JSON-format tagged value e.g. `modemUniqueId` would return
    `(unique_id, modem)`.

    Args:
        tagged_property: The tagged property value, allowing for camelCase.
        tag: Optional to specify the tag, allowing for camelCase tags.
        include_tag: If True, a tuple is returned with the tag as the second
            element.
    
    Returns:
        A string with the original property name, or a tuple with the original
            property value in snake_case, and the tag

    """
    if isinstance(tag, str) and tagged_property.startswith(tag):
        tagged_property = tagged_property.replace(tag, f'{tag}_')
    tagged_property = camel_to_snake(tagged_property)
    parts = tagged_property.split('_', 1)
    if len(parts) > 1:
        tag = parts[0]
        prop = parts[1]
    else:
        tag = None
        prop = parts[0]
    if not include_tag:
        return prop
    return (prop, tag)


def tag_merge(*dicts: dict) -> dict:
    """Merge multiple tagged property dictionaries.
    
    Args:
        *dicts: A set of dictionaries
    
    Returns:
        Merged dictionary with: `{'config':[], 'readOnly': []}`

    """
    merged = { 'config': [], 'readOnly': [] }
    for d in dicts:
        if not isinstance(d, dict) or 'config' not in d or 'readOnly' not in d:
            raise ValueError(f'Invalid tag dictionary {d}')
        merged['config'] = merged['config'] + d['config']
        merged['readOnly'] = merged['readOnly'] + d['readOnly']
    return merged


def json_compatible(obj: object, camel_keys: bool = True) -> dict:
    """Returns a dictionary compatible with `json.dumps` function.

    Nested objects are converted to dictionaries.
    
    Args:
        obj: The source object.
        camel_keys: Flag indicating whether to convert all nested dictionary
            keys to `camelCase`.
        
    Returns:
        A dictionary with nested arrays, dictionaries and other compatible with
            `json.dumps`.

    """
    res = obj
    if camel_keys and isinstance(obj, dict):
        res = {}
        for k, v in obj.items():
            camel_key = snake_to_camel(str(k))
            res[camel_key] = json_compatible(v, camel_keys)
    try:
        json.dumps(res)
    except TypeError:
        try:
            if isinstance(res, list):
                _temp = []
                for element in res:
                    _temp.append(json_compatible(element, camel_keys))
                res = _temp
            if hasattr(res, '__dict__'):
                res = vars(res)
            if isinstance(res, dict):
                res = json_compatible(res, camel_keys)
        except Exception as err:
            _log.error(err)
    finally:
        return res


def equivalent_attributes(reference: object,
                          other: object,
                          exclude: 'list[str]' = None,
                          ) -> bool:
    """Confirms attribute equivalence between objects of the same type.
    
    Args:
        reference: The reference object being compared to.
        other: The object comparing against the reference.
        exclude: Optional list of attribute names to exclude from comparison.
    
    Returns:
        True if all (non-excluded) attribute name/values match.

    """
    if type(reference) != type(other):
        return False
    if not hasattr(reference, '__dict__') or not hasattr(other, '__dict__'):
        return reference == other
    for attr, val in vars(reference).items():
        if exclude is not None and attr in exclude:
            continue
        if not hasattr(other, attr) or val != vars(other)[attr]:
            return False
    return True



"""Operations for interacting with the hostpipe service from a Docker container.

Since commands passed in this way will be run as root, they are preambled by
`runuser -u {user} --` unless they begin with `sudo` in which case `sudo` will
be removed and the command run directly as root.

It assumes the architecture runs from a directory structure where the hostpipe
fifo is called ./hostpipe/pipe relative to the main app launch directory.

References environment variables:

* `HOST_USER` (default: fieldedge)
* `TIMESTAMP_FMT` str (default: YYYY-mm-ddTHH:MM:SS.NNNZ)
* `DEFAULT_TIMEOUT` float (default: 0.25)
* `MAX_FILE_SIZE` int MegaBytes (default 2)

"""
import logging
import os
from datetime import datetime
from logging import DEBUG, Logger
from subprocess import TimeoutExpired, run
from time import sleep, time

_log = logging.getLogger(__name__)

APP_ENV = os.getenv('APP_ENV', 'docker')
HOST_USER = os.getenv('HOST_USER', 'fieldedge')
HOSTPIPE_PATH = os.getenv('HOSTPIPE_PATH', './hostpipe/pipe')
HOSTPIPE_LOG = os.getenv('HOSTPIPE_LOG', './logs/hostpipe.log')
TIMESTAMP_FMT = os.getenv('HOSTPIPE_TS_FMT', 'YYYY-mm-ddTHH:MM:SS.SSSZ')
COMMAND_PREFIX = f'{TIMESTAMP_FMT},[INFO],command='
RESPONSE_PREFIX = f'{TIMESTAMP_FMT},[INFO],result='

DEFAULT_TIMEOUT = float(os.getenv('HOSTPIPE_TIMEOUT', 0.25))
MAX_FILE_SIZE = int(os.getenv('HOSTPIPE_LOGFILE_SIZE', 2)) * 1024 * 1024
HOSTPIPE_LOG_ITERATION_MAX = int(os.getenv('HOSTPIPE_LOG_ITERATION_MAX', 15))
VERBOSE_DEBUG = str(os.getenv('VERBOSE_DEBUG', False)).lower() == 'true'


def host_command(command: str,
                 noresponse: bool = False,
                 timeout: float = DEFAULT_TIMEOUT,
                 pipelog: str = None,
                 test_mode: bool = False,
                 ) -> str:
    """Sends a host command to a pipe (from the Docker container).
    
    The response is read from the hostpipe.log file assuming the fieldedge-core
    script is in place to echo command and response into a log.
    Care should be taken to ensure the timeout is sufficient for the response,
    and the calling function must handle an empty string response.

    Args:
        command: The command to be executed on the host.
        noresponse: Flag if set don't look for a response.
        timeout: The time in seconds to wait for a response (default 0.25).
        pipelog: Override default hostpipe.log, typically used with test_mode.
        test_mode: Boolean to mock responses.
    
    Returns:
        A string with the command response, or empty if no response received.
    
    Raises:
        TimeoutError if hostpipe does not respond within timeout.
        FileNotFoundError if the hostpipe log cannot be found.

    """
    modcommand = _apply_preamble(command)
    command_time = time()
    if not test_mode:
        _log.debug(f'Sending {modcommand} to hostpipe via shell')
        try:
            run(f'echo "{_escaped_command(modcommand)}" > {HOSTPIPE_PATH} &',
                shell=True,
                timeout=timeout)
        except TimeoutExpired:
            err = f'Command {command} timed out waiting for hostpipe'
            _log.error(err)
            raise TimeoutError(err)
    else:
        _log.info(f'test_mode received command: {command}')
    if noresponse:
        return f'{command} sent'
    pipelog = pipelog or HOSTPIPE_LOG
    if not os.path.isfile(pipelog):
        raise FileNotFoundError(f'Could not find file {pipelog}')
    response_str = host_get_response(command,
                                     command_time=command_time,
                                     pipelog=pipelog,
                                     timeout=timeout,
                                     test_mode=test_mode,
                                     ).strip()
    deleted_count = _maintain_pipelog(pipelog)
    if deleted_count > 0:
        _log.info(f'Removed {deleted_count} oldest lines from {pipelog}')
    if _log.getEffectiveLevel() == DEBUG:
        if response_str == '':
            abv_response = '<no response>'
        elif len(response_str) < 25:
            abv_response = response_str.replace('\n', ';')
        else:
            abv_response = response_str[:20].replace("\n", ";") + '...'
        _log.debug(f'Hostpipe: {command} -> {abv_response}')
    return response_str


def _apply_preamble(command: str) -> str:
    if '$HOME' in command:
        command = command.replace('$HOME', f'/home/{HOST_USER}')
    if APP_ENV.lower() != 'docker':
        return command
    preamble = ''
    if command.startswith('sudo '):
        command = command.replace('sudo ', '')
    elif 'runuser' not in command:
        preamble = f'runuser -u {HOST_USER} -- '
    return f'{preamble}{command}'


def _escaped_command(command: str) -> str:
    escaped_command = ''
    for c in command:
        if c == '"':
            escaped_command += r'\\\"'
        else:
            escaped_command += c
    return escaped_command


def _get_line_ts(line: str) -> float:
    iso_time = line.split(',')[0]
    if '.' not in iso_time:
        iso_time = iso_time.replace('Z', '.000Z')
    utc_dt = datetime.strptime(iso_time, '%Y-%m-%dT%H:%M:%S.%fZ')
    return (utc_dt - datetime(1970, 1, 1)).total_seconds()


def host_get_response(command: str,
                      command_time: float = None,
                      pipelog: str = None,
                      timeout: float = DEFAULT_TIMEOUT,
                      test_mode: bool = False,
                      ) -> str:
    """Retrieves the response to the command from the host pipe _log.
    
    Args:
        command: The host command sent previously.
        timeout: The maximum time in seconds to try for a response.
    
    Returns:
        A string concatenating all the response lines following the command.

    Raises:
        FileNotFoundError if the hostpipe log cannot be found.

    """
    calltime = time()
    modcommand = _apply_preamble(command)
    if pipelog is None:
        pipelog = './logs/hostpipe.log'
    if not os.path.isfile(pipelog):
        raise FileNotFoundError(f'Could not find file {pipelog}')
    _log.debug(f'Searching {pipelog} for {modcommand}')
    response = []
    filepass = 0
    while len(response) == 0:
        # test_mode assumes manual step through will usually violate timeout
        if not test_mode and time() > calltime + timeout:
            _log.warning(f'Response to {command} timed out'
                         f' after {timeout} seconds')
            break
        filepass += 1
        if filepass > HOSTPIPE_LOG_ITERATION_MAX:
            _log.warning(f'Exceeded max={HOSTPIPE_LOG_ITERATION_MAX}'
                        ' iterations on hostpipe log')
            break
        if VERBOSE_DEBUG:
            _log.debug(f'{pipelog} read iteration {filepass}')
        lines = open(pipelog, 'r').readlines()
        for line in reversed(lines):
            if (not test_mode and
                command_time is not None and
                _get_line_ts(line) < command_time):
                # older command, skip this pass
                sleep(0.1)
                break
            if ',command=' in line:
                logged_command = line.split(',command=')[1].strip()
                if VERBOSE_DEBUG:
                    _log.debug(f'Found command {logged_command} in {pipelog}'
                               f'({_get_line_ts(line)})'
                               f' with {len(response)} response lines')
                if logged_command != modcommand:
                    # wrong command/response so dump parsed lines so far
                    cts = _get_line_ts(line)
                    to_remove = []
                    for resline in response:
                        rts = _get_line_ts(resline)
                        if rts == cts:
                            to_remove.append(resline)
                    if VERBOSE_DEBUG:
                        _log.debug(f'Mismatch: {logged_command} != {modcommand}'
                                   f' -> dropping {len(to_remove)} response lines')
                    response = [l for l in response if l not in to_remove]
                else:
                    # we reached the original command so can stop parsing response
                    if VERBOSE_DEBUG:
                        _log.debug(f'Found target {modcommand}'
                                   f' with {len(response)} response lines')
                    response = [l[len(RESPONSE_PREFIX):] for l in response]
                    break
            elif ',result=' in line:
                response.append(line)
        if not test_mode:
            sleep(timeout / 2)
    response.reverse()
    return ''.join(response)


def _maintain_pipelog(pipelog: str,
                      max_file_size: int = MAX_FILE_SIZE,
                      test_mode: bool = False,
                      ) -> int:
    """Deletes log entries if over a maximum size and returns the count deleted.

    Deletes both the command and its response.

    Returns the number of lines deleted.
    """
    # TODO: spin a thread to do this in background? or manage in bash/linux
    if not(os.path.isfile(pipelog)):
        raise FileNotFoundError(f'Could not find {pipelog}')
    to_delete = []
    if os.path.getsize(pipelog) > max_file_size:
        lines = open(pipelog, 'r').readlines()
        while os.path.getsize(pipelog) > max_file_size:
            for line in lines:
                if ',result=' in line:
                    to_delete.append(line)
                elif ',command=' in line:
                    if len(to_delete) == 0:
                        to_delete.append(line)
                    else:
                        break
            with open(pipelog, 'w') as f:
                for line in lines:
                    if line not in to_delete:
                        f.write(line)
        if len(to_delete) > 0 and test_mode:
            with open(pipelog, 'w') as f:
                f.writelines(lines)
    return len(to_delete)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fieldedge_utilities']

package_data = \
{'': ['*']}

install_requires = \
['ifaddr>=0.1.7,<0.2.0',
 'paho-mqtt>=1.6.1,<2.0.0',
 'pyserial>=3.5,<4.0',
 'python-dotenv>=0.19.1,<0.20.0']

setup_kwargs = {
    'name': 'fieldedge-utilities',
    'version': '0.14.0',
    'description': 'Utilities package for the FieldEdge project.',
    'long_description': '# Inmarsat FieldEdge Utilities\n\nInmarsat FieldEdge project supports *Internet of Things* (**IoT**) using\nsatellite communications technology.\n\nThis library available on **PyPI** provides:\n\n* A common **`logger`** format and wrapping file facility.\n* A repeating **`timer`** utility (thread) that can be started, stopped,\nrestarted, and interval changed.\n* A simplified **`mqtt`** client that automatically connects\n(by default to a local `fieldedge-broker`).\n* Helper functions for managing files and **`path`** on different OS.\n* An interface for the FieldEdge **`hostpipe`** service for sending host\ncommands from a Docker container, with request/result captured in a logfile.\n* Helper functions **`ip_interfaces`** for finding and validating IP interfaces\nand addresses/subnets.\n* A defined set of common **`protocols`** used for packet analysis and\nsatellite data traffic optimisation.\n* Helpers for **`tag`** assignment of class definitions to expose properties\nfor MQTT transport between microservices, converting between PEP and JSON style.\n* Helpers for managing **`serial`** ports on a host system.\n* Utilities for converting **`timestamp`**s between unix and ISO 8601\n\n[Docmentation](https://inmarsat-enterprise.github.io/fieldedge-utilities/)\n',
    'author': 'geoffbrucepayne',
    'author_email': 'geoff.bruce-payne@inmarsat.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/inmarsat-enterprise/fieldedge-utilities',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""A threaded timer class that allows flexible reconfiguration.

"""
import logging
import threading
from time import time

_log = logging.getLogger(__name__)


class RepeatingTimer(threading.Thread):
    """A repeating thread to call a function, can be stopped/restarted/changed.
    
    Embedded tasks can use threading for continuous repeated operations.  A
    *RepeatingTimer* can be started, stopped, restarted and reconfigured.

    A Thread that counts down seconds using sleep increments, then calls back 
    to a function with any provided arguments.
    Optional auto_start feature starts the thread and the timer, in this case 
    the user doesn't need to explicitly start() then start_timer().

    Attributes:
        name (str): An optional descriptive name for the Thread.
        interval (int): Repeating timer interval in seconds (0=disabled).
        sleep_chunk (float): The fraction of seconds between processing ticks.

    """
    def __init__(self,
                 seconds: int,
                 target: callable,
                 args: tuple = (),
                 kwargs: dict = {},
                 name: str = None,
                 sleep_chunk: float = 0.25,
                 max_drift: int = None,
                 auto_start: bool = False,
                 defer: bool = True,
                 daemon: bool = True,
                 verbose_debug: bool = False,
                 ):
        """Sets up a RepeatingTimer thread.

        Args:
            seconds: Interval for timer repeat.
            target: The function to execute each timer expiry.
            args: Positional arguments required by the target.
            kwargs: Optional keyword arguments to pass into the target.
            name: Optional thread name.
            sleep_chunk: Tick seconds between expiry checks.
            max_drift: Number of seconds clock drift to tolerate.
            auto_start: Starts the thread and timer when created.
            defer: Set if first target waits for timer expiry.
            daemon: Set if thread is a daemon (default)
            verbose_debug: verbose logging of tick count

        Raises:
            ValueError if seconds is not an integer.
        """
        if not (isinstance(seconds, int) and seconds >= 0):
            err_str = 'RepeatingTimer seconds must be integer >= 0'
            raise ValueError(err_str)
        super().__init__(daemon=daemon)
        self.name = name or f'{target.__name__}_timer_thread'
        self.interval = seconds
        if target is None:
            _log.warning(f'No target specified for RepeatingTimer {self.name}')
        self.target = target
        self._exception = None
        self.args = args
        self.kwargs = kwargs
        self.sleep_chunk = sleep_chunk
        self._defer = defer
        self._verbose_debug = verbose_debug
        self._terminate_event = threading.Event()
        self._start_event = threading.Event()
        self._reset_event = threading.Event()
        self._count = self.interval / self.sleep_chunk
        self._timesync = time()
        self.max_drift = max_drift
        if auto_start:
            self.start()
            self.start_timer()

    @property
    def sleep_chunk(self) -> float:
        return self._sleep_chunk

    @sleep_chunk.setter
    def sleep_chunk(self, value: float):
        if 1 % value != 0:
            raise ValueError('sleep_chunk must evenly divide 1 second')
        self._sleep_chunk = value

    @property
    def is_running(self) -> bool:
        return self._start_event.is_set()

    def _resync(self, max_drift: int = None) -> int:
        """Used to adjust the next countdown to account for drift.
        
        NOTE: Untested.
        """
        if max_drift is not None:
            drift = time() - self._timesync % self.interval
            max_drift = 0 if max_drift < 1 else max_drift
            if drift > max_drift:
                _log.warning(f'Detected drift of {drift}s')
                return drift
        return 0

    def run(self):
        """*Note: runs automatically, not meant to be called explicitly.*
        
        Counts down the interval, checking every ``sleep_chunk`` for expiry.
        """
        while not self._terminate_event.is_set():
            while (self._count > 0
                   and self._start_event.is_set()
                   and self.interval > 0):
                if self._verbose_debug:
                    if (self._count * self.sleep_chunk
                        - int(self._count * self.sleep_chunk)
                        == 0.0):
                        #: log debug message at reasonable interval
                        _log.debug(f'{self.name} countdown:'
                                   f' {self._count}'
                                   f' ({self.interval}s'
                                   f' @ step {self.sleep_chunk})')
                if self._reset_event.wait(self.sleep_chunk):
                    self._reset_event.clear()
                    self._count = self.interval / self.sleep_chunk
                self._count -= 1
                if self._count <= 0:
                    try:
                        self.target(*self.args, **self.kwargs)
                        drift_adjust = (self.interval
                                        - self._resync(self.max_drift))
                        self._count = drift_adjust / self.sleep_chunk
                    except BaseException as e:
                        self._exception = e

    def start_timer(self):
        """Initially start the repeating timer."""
        self._timesync = time()
        if not self._defer and self.interval > 0:
            self.target(*self.args, **self.kwargs)
        self._start_event.set()
        if self.interval > 0:
            _log.info(f'{self.name} timer started ({self.interval} s)')
        else:
            _log.warning(f'{self.name} timer cannot trigger (interval=0)')

    def stop_timer(self):
        """Stop the repeating timer."""
        self._start_event.clear()
        _log.info(f'{self.name} timer stopped ({self.interval} s)')
        self._count = self.interval / self.sleep_chunk

    def restart_timer(self):
        """Restart the repeating timer (after an interval change)."""
        if not self._defer and self.interval > 0:
            self.target(*self.args, **self.kwargs)
        if self._start_event.is_set():
            self._reset_event.set()
        else:
            self._start_event.set()
        if self.interval > 0:
            _log.info(f'{self.name} timer restarted ({self.interval} s)')
        else:
            _log.warning(f'{self.name} timer cannot trigger (interval=0)')

    def change_interval(self, seconds: int):
        """Change the timer interval and restart it.
        
        Args:
            seconds (int): The new interval in seconds.
        
        Raises:
            ValueError if seconds is not an integer.

        """
        if (isinstance(seconds, int) and seconds >= 0):
            _log.info(f'{self.name} timer interval changed'
                      f' (old:{self.interval} s new:{seconds} s)')
            self.interval = seconds
            self._count = self.interval / self.sleep_chunk
            self.restart_timer()
        else:
            err_str = 'RepeatingTimer seconds must be integer >= 0'
            _log.error(err_str)
            raise ValueError(err_str)

    def terminate(self):
        """Terminate the timer. (Cannot be restarted)"""
        self.stop_timer()
        self._terminate_event.set()
        _log.info(f'{self.name} timer terminated')
    
    def join(self):
        super(RepeatingTimer, self).join()
        if self._exception:
            raise self._exception
        return self.target
